#include "Graph.h"
#include <boost/program_options.hpp>

using namespace std;

class myComp3D
{
	bool reverse;
	Point3D p;
public:
	myComp3D(Point3D pp, const bool& revparam)
	{
		p = pp;
		reverse = revparam;
	}
	bool operator() (const std::pair<int, Point3D> & lhs, const std::pair<int, Point3D> &rhs) const
	{
		if (reverse) return (computePointDistance(p, lhs.second) > computePointDistance(p, rhs.second));
		else return (computePointDistance(p, lhs.second) < computePointDistance(p, rhs.second));
	}
};

typedef std::priority_queue<std::pair<int, Point3D>, std::vector< std::pair<int, Point3D> >, myComp3D> mypq_type;

/***********************************************/
char * strLicense = "THIS SOFTWARE IS PROVIDED \"AS-IS\". THERE IS NO WARRANTY OF ANY KIND. "
"NEITHER THE AUTHORS NOR THE OHIO STATE UNIVERSITY WILL BE LIABLE FOR "
"ANY DAMAGES OF ANY KIND, EVEN IF ADVISED OF SUCH POSSIBILITY. \n"
"\n"
"This software was developed (and is copyrighted) by Dayu Shi from the Jyamiti"
"group at The Ohio State University. Please do not redistribute this software. "
"This program is for academic research use only. This software uses the "
"Boost library (www.boost.org) "
"which is covered under their own licenses.\n"
"\n"
"The Boost library's license "
"(which applies to the Boost library ONLY and NOT to this program itself) is "
"as follows:\n"
"\n"
"LICENSE\n"
"---------------------------------------------------------------------------\n"
"Boost Software License - Version 1.0 - August 17th, 2003\n"
"\n"
"Permission is hereby granted, free of charge, to any person or organization "
"obtaining a copy of the software and accompanying documentation covered by "
"this license (the \"Software\") to use, reproduce, display, distribute, "
"execute, and transmit the Software, and to prepare derivative works of the "
"Software, and to permit third-parties to whom the Software is furnished to "
"do so, all subject to the following: \n"
"\n"
"The copyright notices in the Software and this entire statement, including "
"the above license grant, this restriction and the following disclaimer, "
"must be included in all copies of the Software, in whole or in part, and "
"all derivative works of the Software, unless such copies or derivative "
"works are solely in the form of machine-executable object code generated by "
"a source language processor. \n"
"\n"
"THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR "
"IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, "
"FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT "
"SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE "
"FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, "
"ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER "
"DEALINGS IN THE SOFTWARE. \n"
"---------------------------------------------------------------------------\n";
/**********************************************************************/
bool ParseCommand(int argc, char **argv, std::string &fileName1, std::string &fileName2, double &delta1, double &delta2, std::string &fileOutput)
{
	try
	{
		/* Define the program options description
		*/
		namespace po = boost::program_options;
		po::options_description desc("PersistenceDistortion(Version 1) Usage");
		desc.add_options()
			(",h", "Help information;")
			(",l", "License information;")
			(",g", po::value<std::string>(&fileName1)->required(), "The file name for the first input graph;")
			(",G", po::value<std::string>(&fileName2)->required(), "The file name for the second input graph;")
			(",d", po::value<double>(&delta1)->required(), "The delta value for sampling the first input graph;")
			(",D", po::value<double>(&delta2)->required(), "The delta value for sampling the second input graph;")
			(",re", po::value<std::string>(&fileOutput)->required(), "The file name for saving the result;");
		// Parser map
		po::variables_map vm;
		try
		{
			po::store(po::parse_command_line(argc, argv, desc), vm);

			//
			if (vm.count("-h"))
			{
				std::cout << desc << std::endl;
			}
			//
			if (vm.count("-l"))
			{
				std::cout << strLicense << std::endl;
			}
			//
			po::notify(vm);
		}
		catch (boost::program_options::required_option& e)
		{
			std::cerr << "ERROR: " << e.what() << std::endl;
			return false;
		}
		catch (boost::program_options::error& e)
		{
			std::cerr << "ERROR: " << e.what() << std::endl;
			return false;
		}
	}
	catch (std::exception& e)
	{
		std::cerr << "Unhandled Exception reached the top of main: "
			<< e.what() << ", application will now exit" << std::endl;
		return false;

	}
	return true;
}

int main(int argc, char **argv)
{
	std::string fileName1, fileName2, fileResult;
	double delta1, delta2;
	if (ParseCommand(argc, argv, fileName1, fileName2, delta1, delta2, fileResult))
	{
		std::ifstream ifs;
		double x, y, z;
		int i1, i2, i3;
		vector<Point3D> vP;
		ifs.open(fileName1.c_str());
		int countP, countF;
		ifs.ignore(256, '\n');
		ifs >> countP >> countF;
		ifs.ignore(256, '\n');
		//reading first graph
		while (countP--)
		{
			ifs >> x >> y >> z;
			Point3D p;
			p.x = x;
			p.y = y;
			p.z = z;
			vP.push_back(p);
		}
		Graph ga(vP.size(), delta1, countF);
		ga.setVP(vP);
		while (countF--)
		{
			ifs >> i1;
			ifs >> i1 >> i2 >> i3;
			ga.addEdge(i1, i2, computePointDistance(vP[i1], vP[i2]));
			ga.addEdge(i2, i3, computePointDistance(vP[i2], vP[i3]));
			ga.addEdge(i1, i3, computePointDistance(vP[i1], vP[i3]));
			ifs.ignore(256, '\n');
		}
		ifs.close();
		ifs.clear();
		//reading second graph
		vP.clear();
		ifs.open(fileName2.c_str());
		ifs.ignore(256, '\n');
		ifs >> countP >> countF;
		ifs.ignore(256, '\n');
		while (countP--)
		{
			ifs >> x >> y >> z;
			Point3D p;
			p.x = x;
			p.y = y;
			p.z = z;
			vP.push_back(p);
		}
		Graph gb(vP.size(), delta2, countF);
		gb.setVP(vP);
		while (countF--)
		{
			ifs >> i1;
			ifs >> i1 >> i2 >> i3;
			gb.addEdge(i1, i2, computePointDistance(vP[i1], vP[i2]));
			gb.addEdge(i2, i3, computePointDistance(vP[i2], vP[i3]));
			gb.addEdge(i1, i3, computePointDistance(vP[i1], vP[i3]));
			ifs.ignore(256, '\n');
		}
		ifs.close();
		ifs.clear();
		//computing persistence-distortion distance
		ga.computePersistenceDistortionDistance(gb, fileResult);
	}
	return 0;
}
